/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "meta/meta.pmh"

using namespace fe;

MetaBrush::MetaBrush(void):
	m_pOperatorContext(NULL)
{
}

MetaBrush::~MetaBrush(void)
{
}

void MetaBrush::initEvent(sp<Scope>& a_rspScope)
{
	m_event.bind(a_rspScope);
	sp<Layout> spLayout=m_event.layout();
	Record record=a_rspScope->createRecord(spLayout);
	m_event.bind(record);
	m_event.reset();

	//* NOTE Houdini may try destruct the brush way too late,
	//* so get rid of references as soon as possible
	FEASSERT(m_pOperatorContext);
	m_spMaster=m_pOperatorContext->master();
	delete(m_pOperatorContext);
	m_pOperatorContext=NULL;
}

void MetaBrush::initDrawInterfaces(MetaPlugin* a_pMetaPlugin)
{
	m_spDrawPerspective=a_pMetaPlugin->drawBrush();
	m_spDrawOrtho=a_pMetaPlugin->drawBrushOverlay();
	updateDrawInterfaces();
}

void MetaBrush::updateDrawInterfaces(void)
{
	sp<DrawCached> spDrawCached=m_spDrawPerspective;
	if(spDrawCached.isValid())
	{
		spDrawCached->setDrawChain(m_spDrawOpenGL);
	}
	spDrawCached=m_spDrawOrtho;
	if(spDrawCached.isValid())
	{
		spDrawCached->setDrawChain(m_spDrawOpenGL);
	}
}

void MetaBrush::setupDrawOpenGL(void)
{
	updateDrawInterfaces();

	sp<DrawMode> spDrawMode(new DrawMode);
	spDrawMode->setDrawStyle(DrawMode::e_solid);
	spDrawMode->setLineWidth(3.0);
	spDrawMode->setPointSize(8.0);
//	spDrawMode->setAntialias(TRUE);

	m_spDrawOpenGL->pushDrawMode(spDrawMode);
}

void MetaBrush::draw(void)
{
	const BWORD createDraw=m_spDrawOpenGL.isNull();
	if(createDraw)
	{
		feLog("MetaBrush::draw creating DrawOpenGL\n");

		if(m_spMaster.isNull())
		{
			feLog("MetaBrush::draw m_spMaster invalid\n");
			return;
		}

		m_spDrawOpenGL=m_spMaster->registry()->create("DrawI.DrawOpenGL");
		m_spMaster=NULL;
	}

	//* sandbox any OpenGL state changes
	m_spDrawOpenGL->pushMatrixState();

	if(createDraw)
	{
		setupDrawOpenGL();
	}

	sp<DrawMode> spDrawMode=m_spDrawOpenGL->popDrawMode();
	m_spDrawOpenGL->pushDrawMode(spDrawMode);

	//* perspective

	sp<ViewI> spViewI=m_spDrawOpenGL->view();
	spViewI->use(ViewMode::e_perspective);

	if(m_spDrawPerspective.isValid())
	{
		m_spDrawOpenGL->pushDrawMode(spDrawMode);
		m_spDrawPerspective->flushLive();
		m_spDrawOpenGL->popDrawMode();
	}

	//* ortho
	sp<ViewMode> spViewMode=spViewI->viewMode();
	spViewMode->setViewport(m_viewport);
	spViewI->use(ViewMode::e_ortho);

	if(m_spDrawOrtho.isValid())
	{
		m_spDrawOpenGL->pushDrawMode(spDrawMode);
		m_spDrawOrtho->flushLive();
		m_spDrawOpenGL->popDrawMode();
	}

	//* restore original state
	m_spDrawOpenGL->popMatrixState();
}
