import sys
import os
import re
forge = sys.modules["forge"]

def setup(module):
	srcList = [	"oiio.pmh",
				"ImageOIIO",
				"oiioDL" ]

	dll = module.DLL( "fexOiioDL", srcList )

	deplibs = forge.basiclibs + [
				"fePluginLib" ]

	incPath = os.path.join(forge.oiio_dist,'include')
	libPath = os.path.join(forge.oiio_dist,'lib')

	oiioMajor = ""
	oiioMinor = ""
	oiioPatch = ""
	for line in open(incPath + "/OpenImageIO/version.h").readlines():
		if re.match("#define OIIO_VERSION_MAJOR.*", line):
			oiioMajor = line.split()[2].rstrip()
		if re.match("#define OIIO_VERSION_MINOR.*", line):
			oiioMinor = line.split()[2].rstrip()
		if re.match("#define OIIO_VERSION_PATCH.*", line):
			oiioPatch = line.split()[2].rstrip()

	version = oiioMajor + "." + oiioMinor + "." + oiioPatch
	if version != "..":
		sys.stdout.write(" " + version)

	module.includemap = { 'oiio' : incPath }

#	dll.linkmap = { "gfxlibs": forge.gfxlibs }
	dll.linkmap = {}

	if forge.fe_os == "FE_LINUX":
		dll.linkmap['oiio'] = "-Xlinker -rpath='" + libPath + "'"
		dll.linkmap["oiio"] += " -L" + libPath + " -lOpenImageIO"
	elif forge.fe_os == "FE_WIN32" or forge.fe_os == "FE_WIN64":
		dll.linkmap["oiio"] = ""	# TODO

	forge.deps( ["fexOiioDLLib"], deplibs )

	forge.tests += [
		("inspect.exe",	"fexOiioDL",						None,	None) ]

	if 'viewer' in forge.module_list and os.path.exists('ext/viewer'):
		forge.tests += [
			("xImage.exe",	"fexOiioDL ../media/image/FreeElectron64.tif 10",
																None,	None),
			("xImage.exe",	"fexOiioDL 10",						None,	None),
			("xRay.exe",	"fexOiioDL 30",						None,	None) ]
	else:
		forge.color_on(0, forge.BLUE)
		sys.stdout.write(" -viewer")
		forge.color_off()

#	module.Module('test')

def auto(module):
	if forge.oiio_dist == '':
		return 'unset'

	test_file = """
#include <OpenImageIO/imagebuf.h>

int main(void)
{
	return(0);
}
    \n"""

	incPath = os.path.join(forge.oiio_dist,'include')
	libPath = os.path.join(forge.oiio_dist,'lib')

	forge.includemap['oiio'] = incPath
#	forge.linkmap['oiio'] = '-L' + libPath + ' -lOpenImageIO'

	result = forge.cctest(test_file)

	forge.includemap['oiio'] = ''
	forge.linkmap['oiio'] = ''

	return result
